@@ -1,5 +1,5 @@
 /*!
-    Copyright (C) 2015-2019 Mateusz Łącki and Michał Startek.
+    Copyright (C) 2015-2020 Mateusz Łącki and Michał Startek.
 
     This file is part of IsoSpec.
 
@@ -16,11 +16,11 @@
 
 #pragma once
 
-#include <tuple>
 #include <unordered_map>
 #include <queue>
 #include <limits>
 #include <string>
+#include <vector>
 #include "platform.h"
 #include "dirtyAllocator.h"
 #include "summator.h"
@@ -28,22 +28,18 @@
 #include "marginalTrek++.h"
 
 
-#if ISOSPEC_BUILDING_R
-#include <Rcpp.h>
-using namespace Rcpp;
-#endif /* ISOSPEC_BUILDING_R */
-
 
 namespace IsoSpec
 {
 
 // This function is NOT guaranteed to be secure against malicious input. It should be used only for debugging.
 unsigned int parse_formula(const char* formula,
-                           std::vector<const double*>& isotope_masses,
-                           std::vector<const double*>& isotope_probabilities,
+                           std::vector<double>& isotope_masses,
+                           std::vector<double>& isotope_probabilities,
                            int** isotopeNumbers,
                            int** atomCounts,
-                           unsigned int* confSize);
+                           unsigned int* confSize,
+                           bool use_nominal_masses = false);
 
 
 //! The Iso class for the calculation of the isotopic distribution.
@@ -51,8 +47,7 @@ unsigned int parse_formula(const char* formula,
     It contains full description of the molecule for which one would like to calculate the isotopic distribution.
 */
 class ISOSPEC_EXPORT_SYMBOL Iso {
-private:
-
+ private:
     //! Set up the marginal isotopic envelopes, corresponding to subisotopologues.
     /*!
         \param _isotopeMasses A table of masses of isotopes of the elements in the chemical formula,
@@ -60,20 +55,21 @@ private:
         \param _isotopeProbabilities A table of isotope frequencies of the elements in the chemical formula,
                                      e.g. {.989212, .010788, .999885, .000115} for C100H202.
     */
-    void setupMarginals(const double* const * _isotopeMasses,
-                        const double* const * _isotopeProbabilities);
+    void setupMarginals(const double* _isotopeMasses,
+                        const double* _isotopeProbabilities);
     bool            disowned;       /*!< A variable showing if the Iso class was specialized by its child-class. If so, then the description of the molecules has been transfered there and Iso is a carcass class, dead as a dodo, an ex-class if you will. */
-protected:
+
+ protected:
     int             dimNumber;      /*!< The number of elements in the chemical formula of the molecule. */
     int*            isotopeNumbers; /*!< A table with numbers of isotopes for each element. */
     int*            atomCounts;     /*!< A table with numbers of isotopes for each element. */
     unsigned int    confSize;       /*!< The number of bytes needed to represent the counts of isotopes present in the extended chemical formula. */
     int             allDim;         /*!< The total number of isotopes of elements present in a chemical formula, e.g. for H20 it is 2+3=5. */
     Marginal**      marginals;      /*!< The table of pointers to the distributions of individual subisotopologues. */
-    double          modeLProb;      /*!< The log-probability of the mode of the isotopic distribution. */
 
-public:
+    bool doMarginalsNeedSorting() const;
 
+ public:
     Iso();
 
     //! General constructror.
@@ -84,6 +80,13 @@ public:
         \param _isotopeMasses A table of tables of masses of isotopes of the elements in the chemical formula, e.g. {{12.0, 13.003355}, {1.007825, 2.014102}} for C100H202.
         \param _isotopeProbabilities A table of tables of isotope frequencies of the elements in the chemical formula, e.g. {{.989212, .010788}, {.999885, .000115}} for C100H202.
     */
+    Iso(
+        int             _dimNumber,
+        const int*      _isotopeNumbers,
+        const int*      _atomCounts,
+        const double*   _isotopeMasses,
+        const double*   _isotopeProbabilities
+    );
     Iso(
         int             _dimNumber,
         const int*      _isotopeNumbers,
@@ -93,14 +96,28 @@ public:
     );
 
     //! Constructor from the formula object.
-    Iso(const char* formula);
+    Iso(const char* formula, bool use_nominal_masses = false);  // NOLINT(runtime/explicit) - constructor deliberately left to be used as a conversion
 
     //! Constructor from C++ std::string chemical formula.
-    inline Iso(const std::string& formula) : Iso(formula.c_str()) {};
+    inline Iso(const std::string& formula, bool use_nominal_masses = false) : Iso(formula.c_str(), use_nominal_masses) {}  // NOLINT(runtime/explicit) - constructor deliberately left to be used as a conversion
+
+    //! Constructor (named) from aminoacid FASTA sequence as C string.
+    /*!
+        \param fasta An aminoacid FASTA sequence. May be upper/lower/mixed case, may contain selenocystein (U). Subisotopologues will be in order: CHNOS, possibly with Se added at an end if present.
+        \use_nominal_masses Whether to use nucleon number instead of the real mass of each isotope during calculations.
+        \add_water Whether the chain should have the terminating -H and -OH groups at the N and C terminus, respectively.
+    */
+    static Iso FromFASTA(const char* fasta, bool use_nominal_masses = false, bool add_water = true);
+
+    //! Constructor (named) from aminoacid FASTA sequence as C++ std::string. See above for details.
+    static inline Iso FromFASTA(const std::string& fasta, bool use_nominal_masses = false, bool add_water = true) { return FromFASTA(fasta.c_str(), use_nominal_masses, add_water); }
 
     //! The move constructor.
     Iso(Iso&& other);
 
+    /* We're not exactly following standard copy and assign semantics with Iso objects, so delete the default assign constructor just in case, so noone tries to use it. Copy ctor declared below. */
+    Iso& operator=(const Iso& other) = delete;
+
     //! The copy constructor.
     /*!
         \param other The other instance of the Iso class.
@@ -125,7 +142,7 @@ public:
     double getMonoisotopicPeakMass() const;
 
     //! Get the log-probability of the mode-configuration (if there are many modes, they share this value).
-    inline double getModeLProb() const { return modeLProb; };
+    double getModeLProb() const;
 
     //! Get the logprobability of the least probable subisotopologue.
     double getUnlikeliestPeakLProb() const;
@@ -136,14 +153,23 @@ public:
     //! Get the theoretical average mass of the molecule.
     double getTheoreticalAverageMass() const;
 
+    //! Get the theoretical variance of the distribution.
+    double variance() const;
+
+    //! Get the standard deviation of the theoretical distribution.
+    double stddev() const { return sqrt(variance()); }
+
     //! Get the number of elements in the chemical formula of the molecule.
-    inline int getDimNumber() const { return dimNumber; };
+    inline int getDimNumber() const { return dimNumber; }
 
     //! Get the total number of isotopes of elements present in a chemical formula.
-    inline int getAllDim() const { return allDim; };
+    inline int getAllDim() const { return allDim; }
 
     //! Add an element to the molecule. Note: this method can only be used BEFORE Iso is used to construct an IsoGenerator instance.
     void addElement(int atomCount, int noIsotopes, const double* isotopeMasses, const double* isotopeProbabilities);
+
+    //! Save estimates of logarithms of target sizes of marginals using Gaussian approximation into argument array. Array priorities must have length equal to dimNumber.
+    void saveMarginalLogSizeEstimates(double* priorities, double target_total_prob) const;
 };
 
 
@@ -153,44 +179,44 @@ public:
 */
 class ISOSPEC_EXPORT_SYMBOL IsoGenerator : public Iso
 {
-protected:
+ public:
+    const double mode_lprob;
+
+ protected:
     double* partialLProbs;  /*!< The prefix sum of the log-probabilities of the current isotopologue. */
     double* partialMasses;  /*!< The prefix sum of the masses of the current isotopologue. */
-    double* partialProbs;/*!< The prefix product of the probabilities of the current isotopologue. */
+    double* partialProbs;   /*!< The prefix product of the probabilities of the current isotopologue. */
 
-public:
+ public:
     //! Advance to the next, not yet visited, most probable isotopologue.
     /*!
         \return Return false if it is not possible to advance.
     */
     virtual bool advanceToNextConfiguration() = 0;
 
-    ISOSPEC_FORCE_INLINE virtual bool advanceToNextConfigurationWithinLayer() { return advanceToNextConfiguration(); };
-    ISOSPEC_FORCE_INLINE virtual bool nextLayer(double) { return false; };
-
     //! Get the log-probability of the current isotopologue.
     /*!
         \return The log-probability of the current isotopologue.
     */
-    virtual double lprob() const { return partialLProbs[0]; };
+    virtual double lprob() const { return partialLProbs[0]; }
 
     //! Get the mass of the current isotopologue.
     /*!
         \return The mass of the current isotopologue.
     */
-    virtual double mass()  const { return partialMasses[0]; };
+    virtual double mass()  const { return partialMasses[0]; }
 
     //! Get the probability of the current isotopologue.
     /*!
         \return The probability of the current isotopologue.
     */
-    virtual double prob() const { return partialProbs[0]; };
+    virtual double prob() const { return partialProbs[0]; }
 
     //! Write the signature of configuration into target memory location. It must be large enough to accomodate it.
     virtual void get_conf_signature(int* space) const = 0;
 
     //! Move constructor.
-    IsoGenerator(Iso&& iso, bool alloc_partials = true);
+    IsoGenerator(Iso&& iso, bool alloc_partials = true);  // NOLINT(runtime/explicit) - constructor deliberately left to be used as a conversion
 
     //! Destructor.
     virtual ~IsoGenerator();
@@ -206,20 +232,23 @@ public:
 */
 class ISOSPEC_EXPORT_SYMBOL IsoOrderedGenerator: public IsoGenerator
 {
-private:
-    MarginalTrek**              marginalResults;            /*!< Table of pointers to marginal distributions of subisotopologues. */
-    std::priority_queue<void*,std::vector<void*>,ConfOrder> pq; /*!< The priority queue used to generate isotopologues ordered by descending probability. */
-    void*                       topConf;                    /*!< Most probable configuration. */
-    DirtyAllocator              allocator;                  /*!< Structure used for alocating memory for isotopologues. */
-    const std::vector<double>** logProbs;                   /*!< Obtained log-probabilities. */
-    const std::vector<double>** masses;                     /*!< Obtained masses. */
-    const std::vector<int*>**   marginalConfs;              /*!< Obtained counts of isotopes. */
-    double                      currentLProb;               /*!< The log-probability of the current isotopologue. */
-    double                      currentMass;                /*!< The mass of the current isotopologue. */
-    double                      currentProb;                /*!< The probability of the current isotopologue. */
+ private:
+    MarginalTrek**              marginalResults;                    /*!< Table of pointers to marginal distributions of subisotopologues. */
+    std::priority_queue<void*, pod_vector<void*>, ConfOrder> pq;   /*!< The priority queue used to generate isotopologues ordered by descending probability. */
+    void*                       topConf;                            /*!< Most probable configuration. */
+    DirtyAllocator              allocator;                          /*!< Structure used for alocating memory for isotopologues. */
+    const pod_vector<double>**  logProbs;                           /*!< Obtained log-probabilities. */
+    const pod_vector<double>**  masses;                             /*!< Obtained masses. */
+    const pod_vector<Conf>**    marginalConfs;                      /*!< Obtained counts of isotopes. */
+    double                      currentLProb;                       /*!< The log-probability of the current isotopologue. */
+    double                      currentMass;                        /*!< The mass of the current isotopologue. */
+    double                      currentProb;                        /*!< The probability of the current isotopologue. */
     int                         ccount;
 
-public:
+ public:
+    IsoOrderedGenerator(const IsoOrderedGenerator& other) = delete;
+    IsoOrderedGenerator& operator=(const IsoOrderedGenerator& other) = delete;
+
     bool advanceToNextConfiguration() override final;
 
     //! Save the counts of isotopes in the space.
@@ -234,7 +263,7 @@ public:
         if (ccount >= 0)
             c[ccount]--;
 
-        for(int ii=0; ii<dimNumber; ii++)
+        for(int ii = 0; ii < dimNumber; ii++)
         {
             memcpy(space, marginalResults[ii]->confs()[c[ii]], isotopeNumbers[ii]*sizeof(int));
             space += isotopeNumbers[ii];
@@ -245,7 +274,7 @@ public:
     };
 
     //! The move-contstructor.
-    IsoOrderedGenerator(Iso&& iso, int _tabSize  = 1000, int _hashSize = 1000);
+    IsoOrderedGenerator(Iso&& iso, int _tabSize  = 1000, int _hashSize = 1000);  // NOLINT(runtime/explicit) - constructor deliberately left to be used as a conversion
 
     //! Destructor.
     virtual ~IsoOrderedGenerator();
@@ -262,8 +291,7 @@ public:
 */
 class ISOSPEC_EXPORT_SYMBOL IsoThresholdGenerator: public IsoGenerator
 {
-private:
-
+ private:
     int*                    counter;            /*!< An array storing the position of an isotopologue in terms of the subisotopologues ordered by decreasing probability. */
     double*                 maxConfsLPSum;
     const double            Lcutoff;            /*!< The logarithm of the lower bound on the calculated probabilities. */
@@ -277,24 +305,30 @@ private:
     double partialLProbs_second_val, lcfmsv;
     bool empty;
 
-public:
+ public:
+    IsoThresholdGenerator(const IsoThresholdGenerator& other) = delete;
+    IsoThresholdGenerator& operator=(const IsoThresholdGenerator& other) = delete;
+
     inline void get_conf_signature(int* space) const override final
     {
         counter[0] = lProbs_ptr - lProbs_ptr_start;
         if(marginalOrder != nullptr)
-            for(int ii=0; ii<dimNumber; ii++)
+        {
+            for(int ii = 0; ii < dimNumber; ii++)
             {
                 int jj = marginalOrder[ii];
                 memcpy(space, marginalResultsUnsorted[ii]->get_conf(counter[jj]), isotopeNumbers[ii]*sizeof(int));
                 space += isotopeNumbers[ii];
             }
+        }
         else
-            for(int ii=0; ii<dimNumber; ii++)
+        {
+            for(int ii = 0; ii < dimNumber; ii++)
             {
                 memcpy(space, marginalResultsUnsorted[ii]->get_conf(counter[ii]), isotopeNumbers[ii]*sizeof(int));
                 space += isotopeNumbers[ii];
             }
-
+        }
     };
 
     //! The move-constructor.
@@ -306,7 +340,7 @@ public:
         \param tabSize The size of the extension of the table with configurations.
         \param hashSize The size of the hash-table used to store subisotopologues and check if they have been already calculated.
     */
-    IsoThresholdGenerator(Iso&& iso, double _threshold, bool _absolute=true, int _tabSize=1000, int _hashSize=1000, bool reorder_marginals = true);
+    IsoThresholdGenerator(Iso&& iso, double _threshold, bool _absolute = true, int _tabSize = 1000, int _hashSize = 1000, bool reorder_marginals = true);
 
     ~IsoThresholdGenerator();
 
@@ -328,7 +362,7 @@ public:
 
         int * cntr_ptr = counter;
 
-        while(idx<dimNumber-1)
+        while(idx < dimNumber-1)
         {
             // counter[idx] = 0;
             *cntr_ptr = 0;
@@ -351,9 +385,9 @@ public:
     }
 
 
-    ISOSPEC_FORCE_INLINE double lprob() const override final { return partialLProbs_second_val + (*(lProbs_ptr)); };
-    ISOSPEC_FORCE_INLINE double mass()  const override final { return partialMasses[1] + marginalResults[0]->get_mass(lProbs_ptr - lProbs_ptr_start); };
-    ISOSPEC_FORCE_INLINE double prob()  const override final { return partialProbs[1] * marginalResults[0]->get_prob(lProbs_ptr - lProbs_ptr_start); };
+    ISOSPEC_FORCE_INLINE double lprob() const override final { return partialLProbs_second_val + (*(lProbs_ptr)); }
+    ISOSPEC_FORCE_INLINE double mass()  const override final { return partialMasses[1] + marginalResults[0]->get_mass(lProbs_ptr - lProbs_ptr_start); }
+    ISOSPEC_FORCE_INLINE double prob()  const override final { return partialProbs[1] * marginalResults[0]->get_prob(lProbs_ptr - lProbs_ptr_start); }
 
     //! Block the subsequent search of isotopologues.
     void terminate_search();
@@ -370,7 +404,7 @@ public:
      * and has undefined results (incl. segfaults) otherwise. */
     size_t count_confs();
 
-private:
+ private:
     //! Recalculate the current partial log-probabilities, masses, and probabilities.
     ISOSPEC_FORCE_INLINE void recalc(int idx)
     {
@@ -384,6 +418,15 @@ private:
         partialLProbs[0] = *partialLProbs_second + marginalResults[0]->get_lProb(counter[0]);
         lcfmsv = Lcutoff - partialLProbs_second_val;
     }
+
+    ISOSPEC_FORCE_INLINE void short_recalc(int idx)
+    {
+        for(; idx > 0; idx--)
+            partialLProbs[idx] = partialLProbs[idx+1] + marginalResults[idx]->get_lProb(counter[idx]);
+        partialLProbs_second_val = *partialLProbs_second;
+        partialLProbs[0] = *partialLProbs_second + marginalResults[0]->get_lProb(counter[0]);
+        lcfmsv = Lcutoff - partialLProbs_second_val;
+    }
 };
 
 
@@ -392,8 +435,7 @@ private:
 
 class ISOSPEC_EXPORT_SYMBOL IsoLayeredGenerator : public IsoGenerator
 {
-private:
-
+ private:
     int*                    counter;            /*!< An array storing the position of an isotopologue in terms of the subisotopologues ordered by decreasing probability. */
     double*                 maxConfsLPSum;
     double currentLThreshold, lastLThreshold;
@@ -406,30 +448,38 @@ private:
     const double** resetPositions;
     double* partialLProbs_second;
     double partialLProbs_second_val, lcfmsv, last_lcfmsv;
+    bool marginalsNeedSorting;
+
+
+ public:
+    IsoLayeredGenerator(const IsoLayeredGenerator& other) = delete;
+    IsoLayeredGenerator& operator=(const IsoLayeredGenerator& other) = delete;
 
-public:
     inline void get_conf_signature(int* space) const override final
     {
         counter[0] = lProbs_ptr - lProbs_ptr_start;
         if(marginalOrder != nullptr)
-            for(int ii=0; ii<dimNumber; ii++)
+        {
+            for(int ii = 0; ii < dimNumber; ii++)
             {
                 int jj = marginalOrder[ii];
                 memcpy(space, marginalResultsUnsorted[ii]->get_conf(counter[jj]), isotopeNumbers[ii]*sizeof(int));
                 space += isotopeNumbers[ii];
             }
+        }
         else
-            for(int ii=0; ii<dimNumber; ii++)
+        {
+            for(int ii = 0; ii < dimNumber; ii++)
             {
                 memcpy(space, marginalResultsUnsorted[ii]->get_conf(counter[ii]), isotopeNumbers[ii]*sizeof(int));
                 space += isotopeNumbers[ii];
             }
-
+        }
     };
 
-    inline double get_currentLThreshold() const { return currentLThreshold; };
+    inline double get_currentLThreshold() const { return currentLThreshold; }
 
-    IsoLayeredGenerator(Iso&& iso, int _tabSize=1000, int _hashSize=1000, bool reorder_marginals = true, double t_prob_hint = 0.99);
+    IsoLayeredGenerator(Iso&& iso, int _tabSize = 1000, int _hashSize = 1000, bool reorder_marginals = true, double t_prob_hint = 0.99);  // NOLINT(runtime/explicit) - constructor deliberately left to be used as a conversion
 
     ~IsoLayeredGenerator();
 
@@ -439,11 +489,11 @@ public:
         {
             if(advanceToNextConfigurationWithinLayer())
                 return true;
-        } while(nextLayer(-2.0));
+        } while(IsoLayeredGenerator::nextLayer(-2.0));
         return false;
     }
 
-    ISOSPEC_FORCE_INLINE bool advanceToNextConfigurationWithinLayer() override final
+    ISOSPEC_FORCE_INLINE bool advanceToNextConfigurationWithinLayer()
     {
         do{
             lProbs_ptr++;
@@ -451,7 +501,7 @@ public:
             if(ISOSPEC_LIKELY(*lProbs_ptr >= lcfmsv))
                 return true;
         }
-        while(carry());
+        while(carry());  // NOLINT(whitespace/empty_loop_body) - cpplint bug, that's not an empty loop body, that's a do{...}while(...) construct
         return false;
     }
 
@@ -478,24 +528,109 @@ public:
         last_lcfmsv = lastLThreshold - partialLProbs_second_val;
     }
 
-    virtual bool nextLayer(double offset) override final;
+    bool nextLayer(double offset);
 
-private:
+ private:
     bool carry();
 };
 
 
 
+class IsoStochasticGenerator : IsoGenerator
+{
+    IsoLayeredGenerator ILG;
+    size_t to_sample_left;
+    const double precision;
+    const double beta_bias;
+    double confs_prob;
+    double chasing_prob;
+    size_t current_count;
+
+ public:
+    IsoStochasticGenerator(Iso&& iso, size_t no_molecules, double precision = 0.9999, double beta_bias = 5.0);
+
+    ISOSPEC_FORCE_INLINE size_t count() const { return current_count; }
+
+    ISOSPEC_FORCE_INLINE double mass() const override final { return ILG.mass(); }
+
+    ISOSPEC_FORCE_INLINE double prob() const override final { return static_cast<double>(count()); }
+
+    ISOSPEC_FORCE_INLINE double lprob() const override final { return log(prob()); }
+
+    ISOSPEC_FORCE_INLINE void get_conf_signature(int* space) const override final { ILG.get_conf_signature(space); }
+
+    ISOSPEC_FORCE_INLINE bool advanceToNextConfiguration() override final
+    {
+        /* This function will be used mainly in very small, tight loops, therefore it makes sense to
+         * aggressively inline it, despite its seemingly large body.
+         */
+        while(true)
+        {
+            double curr_conf_prob_left, current_prob;
+
+            if(to_sample_left <= 0)
+                return false;
+
+            if(confs_prob < chasing_prob)
+            {
+                // Beta was last
+                current_count = 1;
+                to_sample_left--;
+                ILG.advanceToNextConfiguration();
+                current_prob = ILG.prob();
+                confs_prob += current_prob;
+                while(confs_prob <= chasing_prob)
+                {
+                    ILG.advanceToNextConfiguration();
+                    current_prob = ILG.prob();
+                    confs_prob += current_prob;
+                }
+                if(to_sample_left <= 0)
+                    return true;
+                curr_conf_prob_left = confs_prob - chasing_prob;
+            }
+            else
+            {
+                // Binomial was last
+                current_count = 0;
+                ILG.advanceToNextConfiguration();
+                current_prob = ILG.prob();
+                confs_prob += current_prob;
+                curr_conf_prob_left = current_prob;
+            }
+
+            double prob_left_to_1 = precision - chasing_prob;
+            double expected_confs = curr_conf_prob_left * to_sample_left / prob_left_to_1;
+
+            if(expected_confs <= beta_bias)
+            {
+                // Beta mode: we keep making beta jumps until we leave the current configuration
+                chasing_prob += rdvariate_beta_1_b(to_sample_left) * prob_left_to_1;
+                while(chasing_prob <= confs_prob)
+                {
+                    current_count++;
+                    to_sample_left--;
+                    if(to_sample_left == 0)
+                        return true;
+                    prob_left_to_1 = precision - chasing_prob;
+                    chasing_prob += rdvariate_beta_1_b(to_sample_left) * prob_left_to_1;
+                }
+                if(current_count > 0)
+                    return true;
+            }
+            else
+            {
+                // Binomial mode: a single binomial step
+                size_t rbin = rdvariate_binom(to_sample_left, curr_conf_prob_left/prob_left_to_1);
+                current_count += rbin;
+                to_sample_left -= rbin;
+                chasing_prob = confs_prob;
+                if(current_count > 0)
+                    return true;
+            }
+        };
+    }
+};
 
 
-#if !ISOSPEC_BUILDING_R
-
-void printConfigurations(
-    const   std::tuple<double*,double*,int*,int>& results,
-    int     dimNumber,
-    int*    isotopeNumbers
-);
-#endif /* !ISOSPEC_BUILDING_R */
-
-} // namespace IsoSpec
-
+}  // namespace IsoSpec
